;; Generated package description from rt-liberation.el  -*- no-byte-compile: t -*-
(define-package "rt-liberation" "3.0.20220112.132319" "Emacs interface to RT" 'nil :commit "884f16de31451701c513e6b3d6321e56185d4e6c" :authors '(("Yoni Rabkin" . "yrk@gnu.org")) :maintainer '("Yoni Rabkin" . "yrk@gnu.org") :keywords '("rt" "tickets") :url "http://www.nongnu.org/rtliber/")
